export class MatchError extends Error {
  constructor(name) {
    super(`Pattern not matched: ${name}`)
  }
}

const ValueKey = Symbol()

const create = (...args) => Object.freeze(Object.assign(...args))

function constant() { return this }
function extract() { return this[ValueKey] }
function extend(f) { return this.of(f(this)) }
function bind(fn) { return fn(this[ValueKey]) }
function ap(other) { return other.map(this[ValueKey]) }
function map(fn) {
  return this.of(this.bind(fn))
}

function match(pattern) {
  const name = this.of.name

  if (name in pattern) {
    return this.bind(pattern[name])
  } else if ('_' in pattern) {
    return pattern['_']()
  } else {
    throw new MatchError(name)
  }
}

const Value = v => ({ [ValueKey]: v })

const Functor = {
  map
}

const ConstantFunctor = {
  map: constant
}

const ConstantMonad = {
  bind: constant
}

const PointedFunctor = of => ({
  of
})

const Applicative = {
  ...Functor,
  ap
}

const Monad = {
  ...PointedFunctor,
  bind
}

const Comonad = {
  extract,
  extend
}

export const Identity = value => create(
  Value(value),
  Functor,
  PointedFunctor(Identity),
  Monad
)

export const Constant = value => create(
  Value(value),
  ConstantFunctor,
  ConstantMonad,
  PointedFunctor(constant)
)

export const Ok = value => create(
  Value(value),
  Functor,
  PointedFunctor(Ok),
  Monad,
  {
    isOk: true,
    isErr: false,
    match
  })

export const Err = value => create(
  Value(value),
  ConstantFunctor,
  ConstantMonad,
  PointedFunctor(Err),
  {
    isOk: false,
    isErr: true,
    match
  }
)

export const Some = value => create(
  Value(value),
  Functor,
  PointedFunctor(Some),
  Monad,
  {
    isSome: true,
    isNone: false,
    match
  }
)

export const None = () => NoneConstant

const NoneConstant = create(
  ConstantFunctor,
  ConstantMonad,
  PointedFunctor(None),
  {
    isSome: false,
    isNone: true,
    match
  }
)

export function curry(func) {
  return function curried(...args) {
    if (args.length >= func.length) {
      return func.apply(this, args)
    } else {
      return function(...args2) {
        return curried.apply(this, args.concat(args2))
      }
    }
  }
}


